/*
 * Copyright (C) 2000  Transmeta Corporation
 *
 * written by Claudio Fleiner <cfl@transmeta.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/reboot.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>   
#include <linux/fs.h>
#include <sys/ioctl.h>
#include <string.h>

void usage() {
        fprintf(stderr,"usage: upgrade_root [-q] [-p] [-b] [-f] image device\n"
                       "-q: quiet, don't print informational messages\n"
                       "-p: show progress bar\n"
                       "-f: fast (don't check what was written)\n"
                       "-b: reboot the machine once the partition has been updated\n");
        exit(1);
}

static int progress_bar=0;
static int progress_current=0;
static int progress_top=0;
static char progress_char='*';
#define PRLEN 65
static void init_progress_bar(char *title,char c,int top) {
        char b[PRLEN+1];
        if(!progress_bar) return;
        memset(b,'-',65);
        b[PRLEN]=0;
        printf("%-10s: |%s|\r%-10s: |",title,b,title);
        fflush(stdout);
        progress_top=top;
        progress_char=c;
        progress_current=0;
}

static void set_progress_bar(int c) {
        int o,n;
        if(c<=progress_current || !progress_bar) return;
        o=progress_current*PRLEN/progress_top;
        n=c*PRLEN/progress_top;
        if(n<=o) return;
        while(o++<n) putchar(progress_char);
        fflush(stdout);
        progress_current=c;
}

static void reset_progress_bar() {
        printf("\r%79s\r","");
        fflush(stdout);
}


#define BUFLEN 4096
int main(int argc,char *argv[]) {
        FILE *in,*out;
        unsigned int device_blocks;
        unsigned long image_size;
        unsigned long long device_size;
        char buf[BUFLEN];
        char buf2[BUFLEN];
        unsigned int i;
        int fd,c,fast=0;
        int boot=0,quiet=0;

        while(argv[1][0]=='-') {
                if(strcmp(argv[1],"-q")==0) {
                        quiet=1;
                        argv++;
                        argc--;
                } else if(strcmp(argv[1],"-f")==0) {
                        fast=1;
                        argv++;
                        argc--;
                } else if(strcmp(argv[1],"-p")==0) {
                        progress_bar=1;
                        argv++;
                        argc--;
                } else if(strcmp(argv[1],"-b")==0) {
                        argv++;
                        argc--;
                        boot=1;
                } else {
                        usage();
                }
        }

        if(argc!=3) usage();

        in=fopen(argv[1],"r");
        if(in==NULL) {
                fprintf(stderr,"cannot open image %s, aborting",argv[1]);
                exit(1);
        }
        fseek(in,0,SEEK_END);
        image_size=ftell(in);
        fseek(in,0,SEEK_SET);


        fd=open(argv[2],O_WRONLY|O_SYNC,0600);  
        if(fd<0) {
                fprintf(stderr,"cannot open blockdevice %s, aborting\n",argv[2]);
                exit(1);
        }

        out=fdopen(fd,"w");
        if(out==NULL) {
                fprintf(stderr,"cannot open blockdevice %s, aborting\n",argv[2]);
                exit(1);
        }

         if(ioctl(fileno(out),BLKGETSIZE,&device_blocks)) {
                 fprintf(stderr,"cannot determine size of blockdevice\n");
                 exit(1);
         }
         device_size=device_blocks*512;

         if(image_size>device_size) {
                 fprintf(stderr,"image file larger than blockdevice, aborting\n");
                 exit(1);
         }

         if(!quiet) printf("DO NOT INTERRUPT NOW!!!\nwriting %ld bytes\n",image_size);

         mlockall(MCL_CURRENT|MCL_FUTURE); // no paging!
         nice(-20); // no interruptions if possible

         c=0;
         signal(SIGHUP,SIG_IGN);
         signal(SIGINT,SIG_IGN);
         signal(SIGQUIT,SIG_IGN);
         signal(SIGTERM,SIG_IGN);
         signal(SIGALRM,SIG_IGN);
         signal(SIGUSR1,SIG_IGN);
         signal(SIGUSR2,SIG_IGN);
         signal(SIGCHLD,SIG_IGN);

         init_progress_bar("writing",'*',image_size);
         while((i=fread(buf,1,BUFLEN,in))>0) {
                 c+=i;
                 if(fwrite(buf,1,i,out)!=i) {
                         fprintf(stderr,"ohohoh, error while writing, aborting, good luck...\n");
                         exit(1);
                 }
                 set_progress_bar(c);
                 if(!progress_bar) { printf("%d\r",c);fflush(stdout); }
         }
         reset_progress_bar();
         if((unsigned long) c != image_size) {
                 fprintf(stderr,"ayeayeaye, not everything could be written, aborting, good luck...\n");
                 exit(1);
         }

         c=0;
         fclose(out);
         if(!fast) { 
                 out=fopen(argv[2],"r");
                 init_progress_bar("checking",'+',image_size);
                 fseek(in,0,SEEK_SET);
                 while((i=fread(buf,1,BUFLEN,in))>0) {
                         c+=i;
                         if(i!=fread(buf2,1,i,out)) {
                                 // technically this could happen even if there is no
                                 // error, but in practice it does not
                                 fprintf(stderr,"Error while reading the data, probably nothing serious, but aborting anyway\n");
                                 exit(1);
                         }
                         if(strncmp(buf,buf2,i)) {
                                 fprintf(stderr,"read error: the flash is probably bad, aborting.\n");
                                 exit(1);
                         }
                         set_progress_bar(c);
                 }
                 reset_progress_bar();
                 if((unsigned long) c != image_size) {
                         fprintf(stderr,"ayeayeaye, not everything could be read, aborting.\n");
                         exit(1);
                 }
         }

         fclose(in);
         if(!quiet) printf("\nDONE, now syncing...\n");
         sync();
         sleep(2);
         if(boot) {
                 if(!quiet) printf("rebooting...\n");
                 fflush(stdout);
                 reboot(RB_AUTOBOOT);
         }
	 exit(0);
}

